/*
  Reading problem and world configuration files.

  You should not edit this file.  We will use exactly this version of
  it in grading; any modifications you make will be lost.
*/
#include <iostream>
#include <fstream>
#include <list>
#include <string>
#include <dolt-lite.h>
#include <math.h>  // for M_PI
#include "input.h"
#include "simrobot.h"

using namespace std;

void skip_rest_of_line(std::istream &in);

template <class T>
bool read_variable_or_error(std::ifstream &in, const char *name, T &value)
{
  std::string n;
  char equal;
  skip_white_comment(in);
  if(in.eof() || in.bad())
    goto error_mess;
  in >> n;
  if(in.bad() || n != std::string(name))
    goto error_mess;
  skip_white_comment(in);
  in >> equal;
  if(in.bad() || equal != '=')
    goto error_mess;
  in >> value;
  if(in.bad())
    goto error_mess;
  return true;
  
 error_mess:
  std::cerr << "Error in problem file encountered while reading "
	    << " parameter: " << name << std::endl;
  return false;
}

void World::drawGL( const dolt::Graphics &out ) const
{
  b.drawGL(out);
  std::list<dolt::drawablePolygon>::const_iterator i = o.begin();
  for(; i != o.end(); ++i)
    i->drawGL(out);
}

void World::drawPS( const dolt::Graphics &out, dolt::psFile &ps,
		    const dolt::Bbox &bounds, const double &scale ) const
{
  b.drawPS(out, ps, bounds, scale);
  std::list<dolt::drawablePolygon>::const_iterator i = o.begin();
  for(; i != o.end(); ++i)
    i->drawPS(out, ps, bounds, scale);
}

bool World::read_file(const char *file)
{
  std::ifstream in(file);
  if(!in)
    return false;
  
  // read boundary
  skip_white_comment(in);
  if(in.eof() || in.bad())
    return true;
//   std::cout << "about to read boundary" << std::endl;
  in >> b;
//   std::cout << "read boundary" << std::endl;
//   std::cout << " got polygon: " << b << std::endl;
  ////////////////
    // no input operator for lineStrip
//     int n;
//     in >> n; // get number of points
//     for (int k = 0; k < n; k++) {
//       dolt::Point q;
//       in >> q;
//       if(in.bad()) {
// 	std::cerr << "Bad world file " << file << " (reading boundary)"
// 		  << std::endl;
// 	return false;
//       }
//       b.push_back(q);
//     }
    ////////////////

//     b.setFilled(true);
  b.setLineWidth(8);
  b.setColor("gray80");
//     b.setBorderColor("gray10");

    // read obstacles
  dolt::Polygon p;
  while(!in.eof()) {
    p.clear();
    skip_white_comment(in);
    if(in.eof() || in.bad())
      return true;
    in >> p;
    if(in.bad())
      return true;
    o.push_back(p);
    o.back().setFilled(true);
    o.back().setLineWidth(0);
    o.back().setColor("gray40");
  }
  return true;
}

bool Problem::read_file(const char *file, 
			World &w, simRobot &r, CommandList &c)
{
  std::ifstream in(file);
  if(!in)
    return false;
  
  std::string worldfile, sensing, cmdfile;
  double xdsr, xasr, odsr, oasr, rsr;
  if (!read_variable_or_error(in, "world", worldfile)
      || !read_variable_or_error(in, "start-pos", s)
      || !read_variable_or_error(in, "start-ang", a)
      || !read_variable_or_error(in, "ExecutionDstSigmaRatio", xdsr)
      || !read_variable_or_error(in, "ExecutionAngSigmaRatio", xasr)
      || !read_variable_or_error(in, "OdometryDstSigmaRatio", odsr)
      || !read_variable_or_error(in, "OdometryAngSigmaRatio", oasr)
      || !read_variable_or_error(in, "RangeSigmaRatio", rsr)) {
    return false;
  }
  a *= M_PI/180.0; 		// convert to radians

  if (!read_variable_or_error(in, "UseSensing", sensing)) 
    return false;
  else if (sensing == "no")
    g = false;
  else if (sensing == "yes")
    g = true;
  else {
    std::cerr << "Error in problem file: parameter \"UseSensing\""
      " must be either \"yes\" or \"no\"." << std::endl;
    return false;
  }
  if (!read_variable_or_error(in, "commands", cmdfile))
    return false;
  
  if (!w.read_file(worldfile.c_str())) {
    std::cerr << "Error reading world file, giving up" << std::endl;
    return false;
  }
  r = simRobot(&w, xdsr, xasr, odsr, oasr, rsr, s.x(), s.y(), a);


  if (!c.read_file(cmdfile.c_str(), g, false)) {
    std::cerr << "Error reading command file, giving up" << std::endl;
    return false;
  }
  
  return true;
}

// read and store the commands from a file.  if "useSensing" is false
// then ignore any "sense" commands read
bool CommandList::read_file(const char* fname, bool useSensing, 
			    bool verbose)
{
  if (verbose) cout << endl << "Opening file \"" << fname << "\"..." << endl;
  std::ifstream inpFile(fname);
  if (!inpFile)
    return false;

  if (verbose) cout << "Reading commands..." << endl;
  while (true) {
    skip_white_comment(inpFile);
    if (inpFile.eof()) break;
    Command c;
    inpFile >> c;
    if (useSensing || c.type != Command::sense)
      Cmds.push_back(c);
  }
  if (verbose) cout << "End of file reached." << endl;

  current = Cmds.begin();
  return true;
}
  
Command CommandList::getCommand() const 
{ 
  // work around windows const iterator == problem
  vector<Command>::const_iterator const_current = current;
  if (const_current == Cmds.end())
    return Command();
  else 
    return *current;
}

bool CommandList::commandsEnd() const
{ 
  // work around windows const iterator == problem
  vector<Command>::const_iterator const_current = current;

  return const_current == Cmds.end(); 
}


// read a Command from a stream
istream& operator>> (istream& in, Command &c)
{
  string ctype;
  in >> ctype;

  if (ctype == "move") {
    c.type = Command::move;
    in >> c.dist >> c.ang;
    c.ang *= M_PI/180.0;
  }
  else if (ctype == "sense") {
    c.type = Command::sense;
    in >> c.ang;
    c.dist = 0.0;
    c.ang *= M_PI/180.0;
  }
  else 
    cerr << "Unknown command type \"" << ctype 
	 << "\" in istream& operator>> (istream& in, Command& c)" << endl;

  return in;
}

// write a Command to a stream
//
// angle is printed in degrees
ostream& operator<< (ostream& out, const Command &c)
{
  if (c.type == Command::move) {
    out << "move: dist=" <<  c.dist << " ang=" << c.ang*180/M_PI;
  }
  else if (c.type == Command::sense) {
    out << "sense: ang=" << c.ang*180/M_PI;
  }
  else 
    cerr << "Unknown command type \"" << c.type 
	 << "\" in ostream& operator<< (ostream& out, const Command& c)"
	 << endl;

  return out;
}


void skip_rest_of_line(std::istream &in)
{
  in.ignore(1000000, '\n');
}

void skip_white_comment(std::istream &in)
{
  char c;
  while(!in.eof()) {
    in >> c;
    if(c == '#') // comment
      skip_rest_of_line(in);
    else if(c > 0 && !isspace(c)) {
      in.putback(c);
      break;
    }
  }
}
