/*
  Reading problem and world configuration files.

  You should not edit this file.  We will use exactly this version of
  it in grading; any modifications you make will be lost.
*/

#ifndef _INPUT_H
#define _INPUT_H

#include <list>
#include <dolt-lite.h>
#include <iostream>
#include <math.h>
#ifndef M_PI
  #define M_PI 3.1415926535897932384626433833
#endif

class simRobot;
class CommandList;
class World;

void skip_white_comment(std::istream &in);

////////////////////////////////////////////////////////////////

class Problem
{
public:
  Problem() {}

  // accessors for information from the problem description
  inline dolt::Point start_pos() const { return s; }
  inline double start_ang() const { return a; }

  inline bool UseSensing() const { return g; }

  // read the problem description file
  bool read_file(const char *file, 
		 World &world, simRobot &r, CommandList &c);


protected:
  dolt::Point s;
  double a;
  bool g;
  simRobot *r;
};

////////////////////////////////////////////////////////////////

class CommandList;

// a single command.  In a "move" command, "ang" represents the angle
// for a turn.  In a "sense" command, "ang" represents the direction
// the range sensor.  In both cases, "ang" is relative to the robot's
// heading.
// 
class Command {
public:
  enum CommandType {move, sense};
  CommandType type;
  double dist; 			// distance for a "move" command
  double ang;  // for "move" command, angle of turn (in radians)
               // for "sense" command, sensor heading (in radians)
               //   relative to robot heading

  // constructor for a move command
  Command(double d, double a) : type(Command::move), dist(d), ang(a) {}
  // constructor for a sense command
  Command(double a) : type(Command::sense), dist(0.0), ang(a) {}
  // default constructon
  Command() : dist(0.0), ang(0.0) {}
};

// read/write a Command from/to a stream
std::istream& operator>> (std::istream& in, Command& c);
std::ostream& operator<< (std::ostream& out, const Command& c);

// a class to manage the list of commands for the robot
//
// 
class CommandList {
private:
  std::vector<Command> Cmds;
  std::vector<Command>::iterator current;

public:
  // read command file and set current command to be the first
  //
  // if "useSensing" is false, then sensing commands are ignored when
  // reading the file.
  CommandList() {}

  bool read_file(const char* fname, bool useSensing, bool verbose = false);

  // methods to access commands by index (zero indexed)
  int numCommands() const { return int(Cmds.size());}
  Command getCommand(int i) const { return Cmds[i];}

  // methods to access commands sequentially
  Command getCommand() const; 
  Command nextCommand() { return *(current++); }
  Command prevCommand() { return *(current--); }
  bool commandsEnd() const;
  void restartCommands() { current = Cmds.begin(); }
};

////////////////////////////////////////////////////////////////

class World : public dolt::drawableObject
{
public:
  // accessor for the boundary and obstacles
//   inline const dolt::drawableLineStrip & boundary() const
  inline const dolt::drawablePolygon & boundary() const { return b; }
  inline const std::list<dolt::drawablePolygon>& obstacles() const
  { return o; }

  // read in the world description from a file
  bool read_file(const char *file);

  // things required for a drawable object
  void drawGL( const dolt::Graphics &out ) const;
  void drawPS( const dolt::Graphics &out, dolt::psFile &ps,
	       const dolt::Bbox &bounds, const double &scale ) const;

protected:
//   dolt::drawableLineStrip b;
  dolt::drawablePolygon b;
  std::list<dolt::drawablePolygon> o;
};


#endif // _INPUT_H
