////////////////////////////////////////////////////////////////
//
// This file is where you should write your code for the assignment.
//
// Most of the code in this file you will not need to change.  Places
// that you will have to write code are marked with four asterisks,
// i.e., "****".  Note that there is some "dummy" code in all these
// places right now.
// 
////////////////////////////////////////////////////////////////
#include <svl/SVL.h>
#include "input.h"
#include "simrobot.h"
#include <dolt-lite.h>

using namespace dolt;
using namespace std;

////////////////////////////////////////////////////////////////
// global variables (from assign2.cpp)
extern Problem prob;
extern objectList locGraphics;
extern CommandList cmds;
extern simRobot R;

// current estimate and covariance
// 
// these need to be general matrices (instead of vectors) because you
// will be using some non-square matrices in your computations
Mat xhat(3,1);
Mat Px(3,3);

// graphics for displaying information about the robot
drawableLineStrip path;
drawableSegment beam;
drawablePoint trueLoc;
drawableSegment trueOrient;
// graphics for showing the configuration estimate computed by your program
drawablePoint locEstimate;
drawableEllipse locUncertainty;
drawableSegment orientEstimate;
// these two segments comprise a "cone" to show the uncertainty limits
// of the orientation
drawableSegment orientUncertaintyLo, orientUncertaintyHi;

////////////////////////////////////////////////////////////////
// prototypes

void updateLocalizationGraphics(const Mat& x, const Mat& P, bool addPathPoint);
void displayBeam(const Command& c, double range);
void merge_sensing(const Command& c, double range);
void compound_odometry(const Command& c);
Segment orientationSegment(Point loc, double ang);


////////////////////////////////////////////////////////////////
// this is what gets called when the user presses 'n'
void nextLocalizationStep()
{
  static bool beamDisplayed = false;
  static Command c;
  static double sensorReading;

  if (!beamDisplayed) {
    if (cmds.commandsEnd()) {
      cout << endl << "Done executing all robot commands" << endl;
      return;
    }
    c = cmds.nextCommand();
    cout << endl << "Executing command: " << c << endl;
    if (c.type == Command::move) {
      compound_odometry(c);   // move the robot and update "xhat" and "Px"
      updateLocalizationGraphics(xhat, Px, true);
    }
    else {                    // get the sensor reading and display beam
      beamDisplayed = true;
      sensorReading = R.getRange(c.ang);
      cout << "  returned range is " << sensorReading << endl;
      displayBeam(c, sensorReading);
    }
  }
  else {                      // remove beam from display...
    locGraphics.remove(&beam);
    beamDisplayed = false;
    cout << endl << "Updating robot pose..." << endl;
    merge_sensing(c, sensorReading); // update "xhat" and "Px"
    updateLocalizationGraphics(xhat, Px, false);
  }

}


void initLocalization()
{
  // set initial robot state and covariance
  xhat[0][0] = prob.start_pos().x();
  xhat[1][0] = prob.start_pos().y();
  xhat[2][0] = prob.start_ang();
  Px = Mat(3, 3, vl_zero);			// the zero matrix

  // set default properties
  drawableLineObject::setDefaultLineWidth(2.0f);
  drawableEnclosedObject::setDefaultLineWidth(2.0f);
  drawablePoint::setDefaultSize(5.0f);

  // add everything to the objectList
  locGraphics.push_back(&path);
  locGraphics.push_back(&locUncertainty);
  locGraphics.push_back(&trueOrient);
  locGraphics.push_back(&trueLoc);
  locGraphics.push_back(&orientEstimate);
  locGraphics.push_back(&orientUncertaintyLo);
  locGraphics.push_back(&orientUncertaintyHi);
  locGraphics.push_back(&locEstimate);

  // set colors and other properties
  locUncertainty.setFilled(false);
  locUncertainty.setLineWidth(3.0);
  locUncertainty.setColor(Color("ivory"));
  path.setColor(Color("yellow"));
  trueLoc.setSize(6.0);


  updateLocalizationGraphics(xhat, Px, true);
}


void displayBeam(const Command& c, double range)
{
  // calculate the beam endpoint
  Vec2 robot_pos(xhat[0][0], xhat[1][0]);
  Vec2 pos(range, 0.0);
  Vec2 beamEnd = Rot2(c.ang + xhat[2][0])*pos + robot_pos;

  beam = drawableSegment(Point(xhat[0][0], xhat[1][0]),
			 Point(beamEnd[0], beamEnd[1]));
  beam.setColor(Color("white"));

  locGraphics.push_back(&beam);
}

// update the display of localization information using the given
// state estimate "x" and its covariance matrix "P".  It gets the true
// configuration of the robot directly to display this information.
//
// If "addPathPoint" is true, then "x" will be used to add another
// segment to the path.  If false, then "x" will be used to revise the
// last point on the path.
// 
void updateLocalizationGraphics(const Mat& x, const Mat& P, bool addPathPoint)
{
  // update the true location and orientation graphics
  double tx,ty,ta;
  R.getTruePosition(&tx, &ty, &ta);
  trueLoc = Point(tx,ty);
  trueOrient = orientationSegment(trueLoc, ta);

  R.setConfig(xhat[0][0], xhat[1][0], xhat[2][0]);
  locEstimate = Point(xhat[0][0], xhat[1][0]);
  orientEstimate = orientationSegment(locEstimate, xhat[2][0]);
  if (addPathPoint)
    path.push_back(locEstimate);
  else
    path.back() = locEstimate;
  
  // show the uncertainty of the orientation estimate
  // 
  // **** This just uses fixed values of +/- 10 degrees --- you need
  // **** to compute the actual values for the 95.4% confidence bound!
  // 
  double hiAng =  10*M_PI/180.0; 
  double loAng = -10*M_PI/180.0;
  orientUncertaintyHi = orientationSegment(locEstimate, hiAng);
  orientUncertaintyLo = orientationSegment(locEstimate, loAng);

  // set the elliptical confidence limits on the position
  //
  // **** This just uses a random orientation and fixed radii --- you
  // **** need to compute the actual value for the 95.4% confidence
  // **** bound!
  // 
  locUncertainty.setPos(locEstimate);
  locUncertainty.setOrientation(rand()/(RAND_MAX+1.0)*2*M_PI);
  locUncertainty.setMajorRadius(0.2);
  locUncertainty.setMinorRadius(0.1);

  // set colors (these have to be here because the old colors of
  // drawableSegments and drawablePoints are clobbered when the global
  // is assigned a new drawable object)
  orientEstimate.setColor(Color("purple"));
  orientUncertaintyLo.setColor(Color("magenta"));
  orientUncertaintyHi.setColor(Color("magenta"));
  trueOrient.setColor(Color("green"));
  trueOrient.setLineWidth(1.0);
  trueLoc.setColor(Color("green"));
  locEstimate.setColor(Color("orchid"));
}

// return a segment that starts at the point "loc" and is at the
// orientation "ang"
//
Segment orientationSegment(Point loc, double ang)
{
  const float orientationSegmentLength = 0.15;
 
  return Segment(loc, Point(loc.x()+orientationSegmentLength*cos(ang),
			    loc.y()+orientationSegmentLength*sin(ang)));
}

// **** this is where you should move the simulated robot (using the
// **** forward and turn functions) and then update the state estimate
// 
void compound_odometry(const Command& c)
{
  // command the simulated robot to move
  R.forward(c.dist);
  R.turn(c.ang);

  // just as an example, generate a random robot movement to move the
  // estimated location around
  double dx = 0.2*(rand()/(RAND_MAX+1.0));
  double dy = 0.2*(rand()/(RAND_MAX+1.0));
  double dth = M_PI/12*(2*(rand()/(RAND_MAX+1.0)) - 1); // +/- 15 degrees

  xhat[0][0] += dx;
  xhat[1][0] += dy;
  xhat[2][0] += dth;
}

// **** this is where you use the information from the range sensing
// **** to update the state estimate.
// 
void merge_sensing(const Command& c, double range)
{
  // just as an example, generate a random robot movement to move the
  // estimated location around
  double dx = 0.2*(rand()/(RAND_MAX+1.0));
  double dy = 0.2*(rand()/(RAND_MAX+1.0));
  double dth = M_PI/12*(2*(rand()/(RAND_MAX+1.0)) - 1); // +/- 15 degrees

  xhat[0][0] += dx;
  xhat[1][0] += dy;
  xhat[2][0] += dth;
}
