////////////////////////////////////////////////////////////////
//
// simrobot.cc
// Copyright (c) 2002-2003 by Wesley H. Huang.  All rights reserved.
// 
////////////////////////////////////////////////////////////////
#include <iostream>
#include <math.h>
#include "simrobot.h"
#include "input.h"

// for laser rangefinder
//#include <CGAL/Object.h>
// #include <CGAL/Ray_2_Segment_2_intersection.h>

using namespace std;
using namespace dolt;

Robot::Robot(double X, double Y, double Ang)
{
  // set the body color of the robot
  setColor(Color(0.0,0.0,1.0));

  // fixed robot shape
  shape.push_back(Point(-0.15,-0.15));
  shape.push_back(Point(+0.15,-0.10));
  shape.push_back(Point(+0.15,+0.10));
  shape.push_back(Point(-0.15,+0.15));

  // create the points of the polygon
  for (int k=0; k<4; k++)
    this->push_back(Point(0.0, 0.0));

  setConfig(X, Y, Ang);
}

void Robot::setConfig(double X, double Y, double Ang)
{
  x=X;
  y=Y;
  ang=Ang;

  // set the points of the drawablePolygon so that the robot displays
  // at the proper position and orientation on the screen
  const double c = cos(ang);
  const double s = sin(ang);
  for (Polygon::Vertex_iterator k = shape.vertices_begin(),
	 j = this->vertices_begin();
       k != shape.vertices_end(); k++, j++) {
    const Point p(x + c*k->x() - s*k->y(), 
		  y + s*k->x() + c*k->y());
//     this->set(j, p);
    *j = p;
  }
}

void Robot::getConfig(double *X, double *Y, double *Ang)
{
  *X=x;
  *Y=y;
  *Ang=ang;
}


////////////////////////////////////////////////////////////////

simRobot::simRobot(World *W, 
		   double _exDstSigRat, double _exAngSigRat,
		   double _odDstSigRat, double _odAngSigRat,
		   double _ranSigRat, 
		   double X, double Y, double Ang)
  :   Robot(X, Y, Ang)
{
  executionDstSigmaRatio = _exDstSigRat;
  executionAngSigmaRatio = _exAngSigRat;
  odometryDstSigmaRatio = _odDstSigRat;
  odometryAngSigmaRatio = _odAngSigRat;
  rangeSigmaRatio = _ranSigRat;

  w = W;  // store the pointer to the world
  trueX = X;
  trueY = Y;
  trueAng = Ang;
}

double simRobot::forward(double dist)
{
  // simulate the movement
  double executionDistSigma = dist * executionDstSigmaRatio;
  double actualDist = gaussian(dist, executionDistSigma);
  trueX += actualDist*cos(trueAng);
  trueY += actualDist*sin(trueAng);
  // calculate the measured distance
  double odometryDistSigma = actualDist * odometryDstSigmaRatio;
  double measuredDist = gaussian(actualDist, odometryDistSigma);
  return measuredDist;
}

double simRobot::turn(double ang)
{
  // simulate the movement
  double executionAngSigma = ang * executionAngSigmaRatio;
  double actualAng = gaussian(ang, executionAngSigma);
  trueAng += actualAng;
  // calculate the measured distance
  double odometryAngSigma = actualAng * odometryAngSigmaRatio;
  double measuredAng = gaussian(actualAng, odometryAngSigma);
  return measuredAng;
}

void simRobot::getTruePosition(double *x, double *y, double *ang) const
{
  *x = trueX;
  *y = trueY;
  *ang = trueAng;
}



////////////////////////////////////////////////////////////////


bool laserIntersect(const Point& p, const Segment& beam, const Polygon& pg,
		    double* d, Segment *s, Point *ip, bool verbose);

void segmentIntersect(const dolt::Segment& a, const dolt::Segment& b,
		      dolt::Point **pInt, dolt::Segment **sInt,
		      bool open=false);

inline double dist(double dx, double dy)
{
  return ::sqrt(dx*dx + dy*dy);
}

inline double dist(const Point& p, const Point& q) 
{
  double dx = p.x() - q.x();
  double dy = p.y() - q.y();

  return dist(dx, dy);
}


double simRobot::getRange(double ang, bool verbose)
{
  double a,b,c;

  double range = predictRange(ang, trueX, trueY, trueAng, &a, &b, &c, verbose);

  return gaussian(range, range*rangeSigmaRatio);
}




double simRobot::predictRange(double ang, double x, double y, double theta,
		    double *dhdx, double *dhdy, double *dhdth,
		    bool verbose)
{
  Point robotLoc(x, y);
  Bbox b = w->boundary().bbox();
  double world_dia = dist(b.xmax()-b.xmin(), b.ymax()-b.ymin());
  Point beam_dir = Point(cos(ang+theta), sin(ang+theta));
  Segment beam(robotLoc, robotLoc + beam_dir * world_dia);
  if (verbose) cout << "Robot's configuration  is (" << x << ", "
		    << y << ", " << theta*180/M_PI << ")" << endl;
  if (verbose) cout << "Sensing in the direction "
		    << (ang+theta)*180/M_PI << endl;
  if (verbose) cout << "The beam is: " << beam << endl;

  double range, r;
  Segment closest, c;
  Point intPoint, ipt;
  
  bool isIntersection;

  // find the range to the world boundary to initialize "range",
  // "closest", and "intPoint"
  if (verbose) cout << "Checking world boundary." << endl;
  const Polygon &wb = w->boundary();
  isIntersection = laserIntersect(robotLoc, beam, wb, 
				  &range, &closest, &intPoint, verbose);

  // if there isn't an intersection with the world boundary, then the
  // robot must be outside the world boundary!!!
  if (!isIntersection) {
    cerr << "No intersection between the laser and the robot boundary!"
	 << endl
	 << "The robot must be outside the boundary (infinite range)."
	 << endl;
    exit(1);
  }

  if (verbose) cout << "Checking obstacles for a closer intersection." << endl;
  for (list<drawablePolygon>::const_iterator ob = w->obstacles().begin();
       ob != w->obstacles().end(); ob++) {
//   if (verbose) cout << "Trying obstacle \"" << (*ob)->name << "\"" << endl;
    isIntersection = laserIntersect(robotLoc, beam, *ob, 
				    &r, &c, &ipt, verbose);
    if (!isIntersection)
      continue;
    if (r < range) {
      if (verbose) cout << "This is a closer intersection." << endl;
      range = r;
      closest = c;
      intPoint = ipt;
    }
  }

  // now calculate the required derivates
  double x_r = x;		// robot location
  double y_r = y;		// robot location
  double phi = ang + theta;	// sensor direction

  // cgal version
//   double w_x = closest.direction().dx();	// vector along the wall
//   double w_y = closest.direction().dy();	// vector along the wall

  // noncgal version (i don't think this has to be a unit vector, but
  // i'll normalize it anyway)
  double w_len = dist(closest.start(), closest.end());
  double w_x = (closest.end().x() - closest.start().x())/w_len;
  double w_y = (closest.end().y() - closest.start().y())/w_len;

  double x_w = intPoint.x();		// laser-wall intersection point
  double y_w = intPoint.y();		// laser-wall intersection point
  double x_c = x_w;		// point on wall
  double y_c = y_w;		// point on wall

  double dx_wdx_r = (w_x * sin(phi)) / (w_x * sin(phi) - w_y* cos(phi));
  double dx_wdy_r = (- w_x* cos(phi)) / (w_x * sin(phi) - w_y* cos(phi));
  double dx_wdtheta = (x_r * w_x * cos(phi) 
		       + (x_c * w_y - w_x * (y_c - y_r)) * sin(phi)
		       - x_w * (w_x * cos(phi) + w_y  * sin(phi)))
    / (w_x * sin(phi) - w_y * cos(phi));
  double dy_wdx_r = (w_y  * sin(phi)) / (w_x * sin(phi) - w_y* cos(phi));
  double dy_wdy_r = (- w_y * cos(phi)) / (w_x * sin(phi) - w_y* cos(phi));
  double dy_wdtheta = (y_r * w_y  * sin(phi) 
		       - (w_y * (x_c - x_r) - y_c * w_x) * cos(phi)
		       - y_w * ( w_y  * sin(phi) + w_x * cos(phi)))
    / (w_x * sin(phi) - w_y * cos(phi));


  *dhdx = ((x_r - x_w) * (1 - dx_wdx_r) - (y_r - y_w)* dy_wdx_r ) 
    / range;
  *dhdy = ((x_r - x_w) * (-dx_wdy_r) + (y_r - y_w) * (1-dy_wdy_r)) 
    / range;
  *dhdth = -((x_r - x_w) * dx_wdtheta + (y_r - y_w) * dy_wdtheta)
    / range;

  return range;
}

// find the closest intersection point (if any) between the ray "r"
// (which originates at point "p") and the polygon "pg"
//
// If no intersection, returns false.
//
// If there is an intersection
//  - returns true
//  - *d is set to the distance between the intersection point and "p"
//  - the Segment that the beam intersects with is copied to *s
//  - *ip is set to be the intersection point
// 
bool laserIntersect(const Point& p, const Segment& beam , const Polygon& pg,
		    double* d, Segment *s, Point *ip, bool verbose)
{
  double range;
  bool rangeInit = false;

  // variables to be used for checking intersection results
  Point *pt, iPoint;
  Segment *sg, closest;
  
  for (Polygon::Edge_const_iterator ei = pg.edges_begin(); 
       ei != pg.edges_end(); ++ei) {
    segmentIntersect(beam, *ei, &pt, &sg, false);
    if (sg) {
      if (verbose) cout << "  Segment intersection!" << endl;
      double da = dist(p, sg->start());
      double db = dist(p, sg->end());
      if (da < range || !rangeInit) {
	rangeInit = true;
	range = da;
	iPoint = sg->start();
	closest = *ei;
      }
      if (db < range) {
	range = db;
	iPoint = sg->end();
	closest = *ei;
      }
      delete sg;
      continue;
    }
    if (pt) {
      double d = dist(p, *pt);
      if (verbose) cout << "  There is an intersection with the segment: "
			<< *ei << endl << "  The intersection point is: "
			<< pt << endl << "  The distance is " << d << endl;
      if (!rangeInit || d < range) {
	rangeInit = true;
	if (verbose) cout << "  This is the closest point so far..." << endl;
	closest = *ei;
	range = d;
	iPoint = *pt;
      }
      delete pt;
    }
  }
    
  if (rangeInit) {
    *s = closest;
    *d = range;
    *ip = iPoint;
    return true;
  }
  else 
    return false;
}

////////////////////////////////////////////////////////////////
//
// Implementation of a random number generator that draws from a
// gaussian distribution.  The underlying (uniform deviate)
// pseudorandom number generator is the Park and Miller "Minimum
// standard" generator with Bayes-Durham shuffle
//
// Copyright (c) 2001 Wesley H. Huang.  All rights reserved.
//
// #include <math.h>
//#include "rand.h"

long pmRand();
void pmRandSeed(long s);
long pmRand();
long pmRandShuffle();
float myRand();

//
// returns a number drawn from a gaussian distribution with the
// specified mean and standard deviation
// 
float gaussian(float mean, float stdev)
{
  static bool cached = false;
  static float extra;  	// the extra number from a 0 mean unit stdev gaussian

  if (cached) {
    cached = false;
    return extra*stdev + mean;
  }
  else {
    // pick a random point in the unit circle (excluding the origin)
    float a,b,c;
    do {
      a = 2*myRand()-1.0;
      b = 2*myRand()-1.0;
      c = a*a + b*b;
    }
    while (c >= 1.0 || c == 0.0);
    // transform it into two values drawn from a gaussian
    float t = ::sqrt(-2*log(c)/c);
    extra = t*a;
    cached = true;
    return t*b*stdev + mean;
  }
}

// default random generator seed (nothing special about this number
// except that it's not 0)
const long defaultSeed = 18846224; 
long pmRandGenSeed = defaultSeed;

// set the seed (optional)
void pmRandSeed(long s)
{
  if (s == 0) // can't use 0 as a seed
    pmRandGenSeed = defaultSeed;	// use the default
  else
    pmRandGenSeed = s;
}
 
// parameters for the random number generator
const long a = 16807;
const long m = 2147483647;
const long q = m/a;		// 127773
const long r = m % a;		// 2836
const long pmRandMax = m;

// Park and Miller's "Minimal Standard" generator using Schrage's
// method to do the computation in 32 bit arithmetic without overflow.
//
// This is a multiplicative congruential generator: I_{j+1} = a I_j mod m
// using the above constants (which are very carefully picked).
//
// The result is an integer between 1 and m-1 inclusive.
//
// Reference: Numerical Recipes in C, 2nd ed, pg 278.
long pmRand()
{

  long k = pmRandGenSeed/q;
  pmRandGenSeed = a*(pmRandGenSeed - k*q) - r*k;
  if (pmRandGenSeed < 0)
    pmRandGenSeed += m;
  return pmRandGenSeed;
}

// the Bayes-Durham shuffle
long pmRandShuffle()
{
  const int size = 32;
  const long div = 1 + (pmRandMax-1)/size;
  static bool initialized = false;
  static long rn[size];
  static long next;
  int index;

  if (!initialized) {
    for (int j= 0; j < size; j++) // warm up
      pmRand();
    for (int k= 0; k < size; k++) // fill table
      rn[k] =pmRand();
    next = pmRand();
    initialized = true;
  }
  index = next/div;		// find out which table entry to use next
  next = rn[index];		// that random nember determines the next index
  rn[index] = pmRand();		// refill the table slot
  return next;			// return the random number
}
  
// 
// This is a simple but pretty good random number generator.  I am
// providing it so that everyone will have the same sequence of
// pseudorandom numbers which will probably be useful for me for
// testing.
//
// Returns a floating point number in (0.0, 1.0)
float myRand()
{
  return pmRandShuffle()/(pmRandMax + 1.0);
}

void myRandSeed(long s)
{
  pmRandSeed(s);
}
