////////////////////////////////////////////////////////////////
//
// simrobot.h
// Copyright (c) 2002-2004 by Wesley H. Huang.  All rights reserved.
// 
////////////////////////////////////////////////////////////////
#ifndef SIMROBOT_H
#define SIMROBOT_H

#include <dolt-lite.h>

class World;

////////////////
// This robot class is just a drawableObject that has the right shape
// for the robot and can be placed anywhere in the world
// 
class Robot : public dolt::drawablePolygon {
private:
  double x,y,ang; 		// current configuration
  dolt::Polygon shape; // shape in robot frame; origin is reference point,
		       // and positive x axis is forwards
public:
  // "Ang" is in radians
  Robot(double X=0.0, double Y=0.0, double Ang=0.0);
  void setConfig(double X, double Y, double Ang);
  void getConfig(double *X, double *Y, double *Ang);
};

////////////////////////////////////////////////////////////////
// The "simRobot" class is a rudimentary robot simulator.  It is
// ultimately a subclass of drawableObject, so it can be displayed on
// the screen.  However, you must set its location yourself based upon
// the "measured" distance that the motion commands return.
// 
class simRobot : public Robot {
private:    
  double trueX, trueY, trueAng;
  World *w;

public:

  double executionDstSigmaRatio;
  double executionAngSigmaRatio;
  double odometryDstSigmaRatio;
  double odometryAngSigmaRatio;
  double rangeSigmaRatio;

  // (X, Y, Ang) is the initial configuration
  simRobot(World *W, 
	   double executionDstSigmaRatio, double executionAngSigmaRatio,
	   double odometryDstSigmaRatio, double odometryAngSigmaRatio,
	   double rangeSigmaRatio, 
	   double X=0.0, double Y=0.0, double Ang=0.0);
  simRobot() {}

  // go forwards for a "dist" and return the odometry measurement
  double forward(double dist);
  // turn an angle "ang" (in radians) and return the odometry measurement
  double turn(double ang);

  // this is only so that you can plot the point on the screen to see
  // how far off your estimate is.  
  void getTruePosition(double *x, double *y, double *ang) const;

  // This function simulates the laser rangefinder.  It points the
  // sensor in the given direction ("ang" in radians, relative to the
  // robot's forward direction), and measures the distance to the
  // first obstacle along that line.  Gaussian noise is added to the
  // measurement before returning it.
  double getRange(double ang, bool verbose=false);

  // This function is what you use with your Kalman filter.  You have
  // to tell it where you think you are, and it will tell you exactly
  // what the laser rangefinder would return from that configuration
  // without any noise added.  It also gives you the derivatives of H
  // that you need for the Kalman filter.
  double predictRange(double ang, double x, double y, double th,
		      double *dhdx, double *dhdy, double *dhdth,
		      bool verbose=false);
};



////////////////////////////////////////////////////////////////
//
// Implementation of a random number generator that draws from a
// gaussian distribution.  The underlying (uniform deviate)
// pseudorandom number generator is the Park and Miller "Minimum
// standard" generator with Bayes-Durham shuffle
//
// Copyright (c) 2001 Wesley H. Huang.  All rights reserved.
//
////////////////////////////////////////////////////////////////
float gaussian(float mean, float stdev);
void myRandSeed(long s);





#endif
